/*------------ -------------- -------- --- ----- ---   --       -            -
 *  fino's boundary conditions
 *
 *  Copyright (C) 2015 jeremy theler
 *
 *  This file is part of fino.
 *
 *  fino is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  fino is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with wasora.  If not, see <http://www.gnu.org/licenses/>.
 *------------------- ------------  ----    --------  --     -       -         -
 */
#include <gsl/gsl_math.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_blas.h>

#include "fino.h"

#undef  __FUNCT__
#define __FUNCT__ "fino_read_bcs"
int fino_read_bcs(void) {

  physical_entity_t *physical_entity;
  
  // barremos los physical entities y mapeamos cadenas a valores enteros
  LL_FOREACH(wasora_mesh.physical_entities, physical_entity) {
    if (physical_entity->bc_type_string != NULL) {
      if (strcasecmp(physical_entity->bc_type_string, "dirichlet") == 0) {
        physical_entity->bc_type_int = BC_DIRICHLET;
        
      } else if (strcasecmp(physical_entity->bc_type_string, "neumann") == 0) {
        physical_entity->bc_type_int = BC_NEUMANN;
        
      } else if (strcasecmp(physical_entity->bc_type_string, "robin") == 0) {
        physical_entity->bc_type_int = BC_ROBIN;
        
      } else if (strcasecmp(physical_entity->bc_type_string, "point_force") == 0) {
        physical_entity->bc_type_int = BC_POINT_FORCE;

      } else {
        wasora_push_error_message("unknown boundary condition type '%s'", physical_entity->bc_type_string);
        return WASORA_PARSER_ERROR;
      }
    }
  }
  
  return WASORA_RUNTIME_OK;
}

#undef  __FUNCT__
#define __FUNCT__ "fino_count_bc_expressions"
int fino_count_bc_expressions(expr_t *bc_args) {
  expr_t *bc_arg;
  int n = 0;
  
  LL_FOREACH(bc_args, bc_arg) {
    n++;
  }
  
  return n;
}

#undef  __FUNCT__
#define __FUNCT__ "fino_evaluate_bc_expressions"
int fino_evaluate_bc_expressions(physical_entity_t *physical_entity, node_t *node, int degrees, double multiplier, double *result) {

  int g;
  expr_t *bc_arg = physical_entity->bc_args;

  // las cc dependen de la posicion
  wasora_var(wasora_mesh.x) = node->x[0];
  wasora_var(wasora_mesh.y) = node->x[1];
  wasora_var(wasora_mesh.z) = node->x[2];

  for (g = 0; g < degrees; g++) {
    if (bc_arg == NULL) {
      wasora_push_error_message("boundary condition '%s' needs %d expressions", physical_entity->name, degrees);
      return WASORA_RUNTIME_ERROR;
    }

    result[g] = multiplier * wasora_evaluate_expression(bc_arg);
    bc_arg = bc_arg->next;
  }
  
  return WASORA_RUNTIME_OK;
}

#undef  __FUNCT__
#define __FUNCT__ "fino_set_essential_bc"
int fino_set_essential_bc(void) {

  int i, j;
  double *phicc = malloc(fino.degrees * sizeof(double));
  if (wasora_var(fino.vars.dirichlet_diagonal) == 0) {
    wasora_var(fino.vars.dirichlet_diagonal) = 1;
  }
  
  for (i = 0; i < fino.mesh->n_elements; i++) {
    if (fino.mesh->element[i].type->dim < fino.dimensions) {
      if (fino.mesh->element[i].type->dim == fino.dimensions && fino.mesh->element[i].physical_entity == NULL) {
        wasora_push_error_message("surface element %d does not have any associated physical entity", i);
        return WASORA_RUNTIME_ERROR;
        
      } else if (fino.mesh->element[i].physical_entity->bc_type_int == BC_UNDEFINED) {
        wasora_push_error_message("physical entity '%s' (id = %d) does not have any associated boundary condition", fino.mesh->element[i].physical_entity->name, fino.mesh->element[i].physical_entity->id);
        return WASORA_RUNTIME_ERROR;
        
      } else if (fino.mesh->element[i].physical_entity->bc_type_int == BC_DIRICHLET) {

        for (j = 0; j < fino.mesh->element[i].type->nodes; j++) {
          MatZeroRows(fino.A, fino.degrees, fino.mesh->element[i].node[j]->index, wasora_var(fino.vars.dirichlet_diagonal), PETSC_NULL, PETSC_NULL);
          wasora_call(fino_evaluate_bc_expressions(fino.mesh->element[i].physical_entity, fino.mesh->element[i].node[j], fino.degrees, wasora_var(fino.vars.dirichlet_diagonal), phicc));
          
          if (fino.problem_type == problem_linear) {
            VecSetValues(fino.b, fino.degrees, fino.mesh->element[i].node[j]->index, phicc, INSERT_VALUES);
          } else if (fino.problem_type == problem_eigen) {
            MatZeroRows(fino.B, fino.degrees, fino.mesh->element[i].node[j]->index, 0.0, PETSC_NULL, PETSC_NULL);
          }
        }
      }
    }
  }
  free(phicc);

  wasora_call(fino_assembly());

  return WASORA_RUNTIME_OK;

}

#undef  __FUNCT__
#define __FUNCT__ "fino_build_surface_objects"
// las condiciones de contorno son
// dphi/dn = a * phi + b
// para neumann, a = 0
int fino_build_surface_objects(element_t *element, expr_t *bc_a, expr_t *bc_b) {
  int v, g;
  double w_gauss;
  expr_t *current_bc_a;
  expr_t *current_bc_b;
  gsl_vector *Nb;
  gsl_matrix *Na;
  gsl_matrix *NaH;
  double *cc;

  if (element->type->dim == 0) {

    // si es un punto es un punto y chau
    cc = malloc(fino.degrees * sizeof(double));
    
    wasora_call(fino_evaluate_bc_expressions(element->physical_entity, element->node[0], fino.degrees, 1, cc));
    VecSetValues(fino.b, fino.degrees, element->node[0]->index, cc, INSERT_VALUES);
    
    free(cc);
    
  } else if (element->type->dim == 1 && fino.dimensions == 3) {
    // las expresiones estan dadas en "lo que sea" por unidad de longitud, asi que le damos la mitad a cada nodo
    double halflength = 0.5 * gsl_hypot3(element->node[1]->x[0] - element->node[0]->x[0],
                                         element->node[1]->x[1] - element->node[0]->x[1],
                                         element->node[1]->x[2] - element->node[0]->x[2]);
    // si es una linea y el problema es 3 hacemos como si fuesen dos puntos
    cc = malloc(fino.degrees * sizeof(double));

    wasora_call(fino_evaluate_bc_expressions(element->physical_entity, element->node[0], fino.degrees, halflength, cc));
    VecSetValues(fino.b, fino.degrees, element->node[0]->index, cc, INSERT_VALUES);

    wasora_call(fino_evaluate_bc_expressions(element->physical_entity, element->node[1], fino.degrees, halflength, cc));
    VecSetValues(fino.b, fino.degrees, element->node[1]->index, cc, INSERT_VALUES);
    
    free(cc);
    
  } else {
    // sino hacemos la cuenta general
  
    if (fino.n_local_nodes != element->type->nodes) {
      wasora_call(fino_allocate_elemental_objects(element));
    }
  
    Na = gsl_matrix_calloc(fino.degrees, fino.degrees);
    NaH = gsl_matrix_calloc(fino.degrees, fino.n_local_nodes);
    Nb = gsl_vector_calloc(fino.degrees);
  
    gsl_matrix_set_zero(fino.Ai);
    gsl_vector_set_zero(fino.bi);

    for (v = 0; v < fino.mesh->n_gauss_surface; v++) {
      w_gauss = mesh_compute_fem_objects_at_gauss(fino.mesh, element, v);    
      current_bc_a = bc_a;
      current_bc_b = bc_b;
      for (g = 0; g < fino.degrees; g++) {
        if (current_bc_a != NULL) {
          gsl_matrix_set(Na, g, g, -wasora_evaluate_expression(current_bc_a));
        }
        gsl_vector_set(Nb, g, wasora_evaluate_expression(current_bc_b));

        if (current_bc_a != NULL) {
          current_bc_a = current_bc_a->next;
        }
        if (current_bc_b != NULL) {
          current_bc_b = current_bc_b->next;
        }
      }

      if (bc_a != NULL) {
        gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1, Na, fino.mesh->fem.H, 0, NaH);
        gsl_blas_dgemm(CblasTrans, CblasNoTrans, w_gauss, fino.mesh->fem.H, NaH, 1, fino.Ai);
      }
    
      gsl_blas_dgemv(CblasTrans, -w_gauss, fino.mesh->fem.H, Nb, 1.0, fino.bi); 
    }
    
    MatSetValues(fino.A, fino.elemental_size, fino.mesh->fem.l, fino.elemental_size, fino.mesh->fem.l, gsl_matrix_ptr(fino.Ai, 0, 0), ADD_VALUES);
    VecSetValues(fino.b, fino.elemental_size, fino.mesh->fem.l, gsl_vector_ptr(fino.bi, 0), ADD_VALUES);
    
    gsl_vector_free(Nb);
    gsl_matrix_free(Na);
    gsl_matrix_free(NaH);
  }
  
  return WASORA_RUNTIME_OK;
}
