/*------------ -------------- -------- --- ----- ---   --       -            -
 *  fino's routines for computing gradients of results
 *
 *  Copyright (C) 2015 jeremy theler
 *
 *  This file is part of fino.
 *
 *  fino is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  fino is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with wasora.  If not, see <http://www.gnu.org/licenses/>.
 *------------------- ------------  ----    --------  --     -       -         -
 */
#include "fino.h"

#undef  __FUNCT__
#define __FUNCT__ "fino_compute_gradients"
int fino_compute_gradients(void) {
  
  double xi;
  int i, j, d, g;
  
  if (fino.mesh->n_cells == 0) {
    mesh_element2cell(fino.mesh);
  }

  // reseteamos (de paso alocamos el data_argument y el data_value)
  // en verdad de los data argument solo alocamos el del primero y hacemos apuntar los otros a este
  fino.gradsol[0][0]->data_argument = calloc(fino.dimensions, sizeof(double *));
  for (d = 0; d < fino.dimensions; d++) {
    fino.gradsol[0][0]->data_argument[d] = calloc(fino.mesh->n_cells, sizeof(double));
  }

  for (g = 0; g < fino.degrees; g++) {
    for (d = 0; d <fino.dimensions; d++) {
      if (!(g == 0 && d == 0)) {
        fino.gradsol[g][d]->data_argument = fino.gradsol[0][0]->data_argument;
      }

      free(fino.gradsol[g][d]->data_value);
      fino.gradsol[g][d]->data_value = calloc(fino.mesh->n_cells, sizeof(double));
      fino.gradsol[g][d]->data_size = fino.mesh->n_cells;

    }
  }

  for (i = 0; i < fino.mesh->n_cells; i++) {
    for (j = 0; j < fino.mesh->cell[i].element->type->nodes; j++) {

      gsl_vector_set_zero(fino.mesh->fem.r);   // en el punto de gauss cero
      mesh_compute_x(fino.mesh->cell[i].element, fino.mesh->fem.r, fino.mesh->fem.x);
      mesh_compute_dxdr(fino.mesh->cell[i].element, fino.mesh->fem.r, fino.mesh->fem.dxdr);
      mesh_inverse(fino.mesh->dimensions, fino.mesh->fem.dxdr, fino.mesh->fem.drdx);
      mesh_compute_dhdx(fino.mesh->cell[i].element, fino.mesh->fem.r, fino.mesh->fem.drdx, fino.mesh->fem.dhdx);


      for (g = 0; g < fino.degrees; g++) {
        for (d = 0; d <fino.dimensions; d++) {

          if (g == 0) {
            fino.gradsol[0][0]->data_argument[d][i] = fino.mesh->cell[i].x[d];
          }

          VecGetValues(fino.phi, 1, &fino.mesh->cell[i].element->node[j]->index[g], &xi);
          fino.gradsol[g][d]->data_value[i] += gsl_matrix_get(fino.mesh->fem.dhdx, j, d) * xi;
        }
      }
    }
  }
  
  return WASORA_RUNTIME_OK;
}
