/*------------ -------------- -------- --- ----- ---   --       -            -
 *  wasora's mesh-related cell routines
 *
 *  Copyright (C) 2014 jeremy theler
 *
 *  This file is part of wasora.
 *
 *  wasora is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  wasora is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with wasora.  If not, see <http://www.gnu.org/licenses/>.
 *------------------- ------------  ----    --------  --     -       -         -
 */
#include <wasora.h>
#include "mesh.h"

int mesh_element2cell(mesh_t *mesh) {
  
  int i_element, i_cell, i_dim, i_node, g, index;
  
  // contamos cuantas celdas hay
  // una celda = un elemento de la dimension del problema
  mesh->n_cells = 0;
  for (i_element = 0; i_element < mesh->n_elements; i_element++) {
    if (mesh->element[i_element].type != NULL && mesh->element[i_element].type->dim == mesh->dimensions) {
      mesh->n_cells++;
    }
  }
  // alocamos las celdas
  mesh->cell = calloc(mesh->n_cells, sizeof(cell_t));
  

  i_cell = 0;
  index = 0;
  for (i_element = 0; i_element < mesh->n_elements; i_element++) {
    if (mesh->element[i_element].type != NULL && mesh->element[i_element].type->dim == mesh->dimensions) {
  
      // las id las empezamos en uno
      mesh->cell[i_cell].id = i_cell+1;
      
      // elemento (ida y vuelta)
      mesh->cell[i_cell].element = &mesh->element[i_element];
      mesh->element[i_element].icell = mesh->cell[i_cell].id;
    
      // holder para vecinos
      mesh->cell[i_cell].ineighbor = calloc(mesh->element[i_element].type->faces, sizeof(int));

      // coordenadas del centro
      for (i_dim = 0; i_dim < mesh->dimensions; i_dim++) {
        mesh->cell[i_cell].x[i_dim] = 0;
        for (i_node = 0; i_node < mesh->cell[i_cell].element->type->nodes; i_node++) {
          mesh->cell[i_cell].x[i_dim] += mesh->cell[i_cell].element->node[i_node]->x[i_dim];
        }
        mesh->cell[i_cell].x[i_dim] /= (double)(mesh->cell[i_cell].element->type->nodes);
      }
      
      // indices globales
      if (mesh->degrees_of_freedom != 0 && mesh->cell[i_cell].index == NULL) {
        mesh->cell[i_cell].index = malloc(mesh->degrees_of_freedom * sizeof(int));
        for (g = 0; g < mesh->degrees_of_freedom; g++) {
          mesh->cell[i_cell].index[g] = index++;
        }
      }

      i_cell++;

    }
  }

  return WASORA_RUNTIME_OK;
  
}


int mesh_compute_coords(mesh_t *mesh) {

  int i, j;
  int index;
  
  index = 0;
  for (i = 0; i < mesh->n_cells; i++) {

    // holder para indices y resultados
    if (mesh->degrees_of_freedom != 0 && mesh->cell[i].index == NULL) {
      mesh->cell[i].index = calloc(mesh->degrees_of_freedom, sizeof(int));
      for (j = 0; j < mesh->degrees_of_freedom; j++) {
        mesh->cell[i].index[j] = index++;
      }
    }
    
    mesh->cell[i].x[0] = mesh->cell[i].x[1] = mesh->cell[i].x[2] = 0;
    // coordenadas del baricentro de la celda = promedio de las coordenadas de los nodos del elementos
    for (j = 0; j < mesh->cell[i].element->type->nodes; j++) {
      mesh->cell[i].x[0] += mesh->cell[i].element->node[j]->x[0];
      mesh->cell[i].x[1] += mesh->cell[i].element->node[j]->x[1];
      mesh->cell[i].x[2] += mesh->cell[i].element->node[j]->x[2];
    }
    mesh->cell[i].x[0] /= (double)(mesh->cell[i].element->type->nodes);
    mesh->cell[i].x[1] /= (double)(mesh->cell[i].element->type->nodes);
    mesh->cell[i].x[2] /= (double)(mesh->cell[i].element->type->nodes);
    
    // esto seria mas lindo con alguna clase de encapsulacion
    mesh->cell[i].volume = mesh->cell[i].element->type->element_volume(mesh->cell[i].element);
    

/*
    if (mesh->dimensions == 1) {
      mesh->cell[i].x[0] = mesh->cell[i].x[1] = mesh->cell[i].x[2] = 0;
    // coordenadas del baricentro de la celda = promedio de las coordenadas de los nodos del elementos
      for (j = 0; j < mesh->cell[i].element->type->nodes; j++) {
        mesh->cell[i].x[0] += mesh->cell[i].element->node[j]->x[0];
        mesh->cell[i].x[1] += mesh->cell[i].element->node[j]->x[1];
        mesh->cell[i].x[2] += mesh->cell[i].element->node[j]->x[2];
      }
      mesh->cell[i].x[0] /= (double)(mesh->cell[i].element->type->nodes);
      mesh->cell[i].x[1] /= (double)(mesh->cell[i].element->type->nodes);
      mesh->cell[i].x[2] /= (double)(mesh->cell[i].element->type->nodes);
      
    } else if (mesh->dimensions == 2) {
      double xi;

      // esto viene de https://en.wikipedia.org/wiki/Centroid#Centroid_of_polygon
      mesh->cell[i].x[0] = mesh->cell[i].x[1] = mesh->cell[i].x[2] = 0;
      for (k = 0; k < mesh->cell[i].element->type->nodes-1; k++) {
        xi = (mesh->cell[i].element->node[k]->x[0]*mesh->cell[i].element->node[k+1]->x[1] - mesh->cell[i].element->node[k+1]->x[0]*mesh->cell[i].element->node[k]->x[1]);
        mesh->cell[i].volume += xi;
        mesh->cell[i].x[0] += (mesh->cell[i].element->node[k]->x[0] + mesh->cell[i].element->node[k+1]->x[0]) * xi;
        mesh->cell[i].x[1] += (mesh->cell[i].element->node[k]->x[1] + mesh->cell[i].element->node[k+1]->x[1]) * xi;
      }
      
      xi = (mesh->cell[i].element->node[k]->x[0]*mesh->cell[i].element->node[0]->x[1] - mesh->cell[i].element->node[0]->x[0]*mesh->cell[i].element->node[k]->x[1]);
      mesh->cell[i].volume += xi;
      mesh->cell[i].x[0] += (mesh->cell[i].element->node[k]->x[0] + mesh->cell[i].element->node[0]->x[0]) * xi;
      mesh->cell[i].x[1] += (mesh->cell[i].element->node[k]->x[1] + mesh->cell[i].element->node[0]->x[1]) * xi;

      mesh->cell[i].volume /= 2;
      mesh->cell[i].x[0] /= 6*mesh->cell[i].volume;
      mesh->cell[i].x[1] /= 6*mesh->cell[i].volume;
      
      // hasta aca el volumen (area) tiene signo segun sea clockwise o counter-clockwise
      mesh->cell[i].volume = fabs(mesh->cell[i].volume);
    }
 */
    
  }
  
  return WASORA_RUNTIME_OK;
}
