/*------------ -------------- -------- --- ----- ---   --       -            -
 *  wasora's mesh-related initialization routines
 *
 *  Copyright (C) 2014 jeremy theler
 *
 *  This file is part of wasora.
 *
 *  wasora is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  wasora is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with wasora.  If not, see <http://www.gnu.org/licenses/>.
 *------------------- ------------  ----    --------  --     -       -         -
 */
#include <wasora.h>
#include "mesh.h"


int wasora_mesh_init_before_parser(void) {
  
  // redireccionamos estos apuntadores
  wasora.interpolate_function_node = mesh_interpolate_function_node;
  wasora.interpolate_function_cell = mesh_interpolate_function_cell;
  wasora.interpolate_function_property = mesh_interpolate_function_property;
  
  // vector con las coordenadas globales
  wasora_mesh.vec_x = wasora_define_vector("x_global", 3, NULL);
  // hacemos alias de a los elementos del vector x_global
  
//va+x+name x
//va+x+desc Holder variable for spatial dependance of functions, such spatial distribution
//va+x+desc of physical properties or results of partial differential equations.
  wasora_mesh.x = wasora_define_variable("x");
  wasora_realloc_variable_ptr(wasora_mesh.x, gsl_vector_ptr(wasora_value_ptr(wasora_mesh.vec_x), 0), 0);

//va+y+name y
//va+y+desc Idem as `x`.
  wasora_mesh.y = wasora_define_variable("y");
  wasora_realloc_variable_ptr(wasora_mesh.y, gsl_vector_ptr(wasora_value_ptr(wasora_mesh.vec_x), 1), 0);

//va+z+name z
//va+z+desc Idem as `x`.
  wasora_mesh.z = wasora_define_variable("z");
  wasora_realloc_variable_ptr(wasora_mesh.z, gsl_vector_ptr(wasora_value_ptr(wasora_mesh.vec_x), 2), 0);

// y ya que estamos las ponemos en un array numerico
  wasora_mesh.arr_x[0] = wasora_mesh.x;
  wasora_mesh.arr_x[1] = wasora_mesh.y;
  wasora_mesh.arr_x[2] = wasora_mesh.z;
  
//va+eps+name eps
//va+eps+desc Small value. Default is $10^{-6}$.
  wasora_mesh.eps = wasora_define_variable("eps");
  wasora_var(wasora_mesh.eps) = 1e-6;
    
  return WASORA_RUNTIME_OK;
}
